:root {
    --pst-color-link: var(--pst-color-primary);
    --pst-color-link-hover: var(--pst-color-secondary);
    --sd-color-primary: var(--pst-color-primary);
    --sd-color-primary-text: var(--pst-color-text-base);
    --sd-color-secondary: #ee9040;
    --sd-color-success: #28a745;
    --sd-color-dark: #323232;
    --sd-color-danger: #dc3545;
    --sd-color-light: #c9c9c9;
}

.simple li>p {
    margin: 0;
}

/* multi column TOC */
.contents ul {
    list-style-type: none;
    padding-left: 2em;
}

.contents > ul {
    padding-left: 0;
}

.multicol-toc > ul {
    column-width: 250px;
    column-gap: 60px;
    -webkit-column-width: 250px;
    -moz-column-width: 250px;
    column-rule: 1px solid #ccc;
}

.multicol-toc > li {
       /* break inside is not yet broadly supported, but we just try */
    break-inside: avoid-column;
    -moz-break-inside: avoid-column;
    -webkit-break-inside: avoid-column;
}

.contents > ul > li > a {
    font-size: 1.0em;
}

/* Hide red ¶ between the thumbnail and caption in gallery

Due the way that sphinx-gallery floats its captions the perma-link
does not float with it.
*/
.sphx-glr-thumbcontainer p.caption:hover > a.headerlink{
    visibility: hidden;
}

/* slightly reduce horizontal margin compared to gallery.css to
 * get four columns of thumbnails in the pydata-sphinx-theme. */
.sphx-glr-thumbcontainer {
    margin: 5px 2px;
}

html[data-theme="dark"] .sphx-glr-thumbcontainer {
    background-color: rgb(63, 63, 63);
}

/* Set a fixed height so that lazy loading does not change heights. Without a fixed
 * height lazy loading of images interferes with anchor links: Clicking a link goes to
 * a certain position, but then the loaded images add content and move the anchor to a
 * different position.
 */
.sphx-glr-thumbcontainer img {
    height: 112px;
}

/* hide download buttons in example headers
 * https://sphinx-gallery.github.io/stable/advanced.html#hide-the-download-buttons-in-the-example-headers
 */
div.sphx-glr-download-link-note {
    display: none;
}

/* re-style the download button */
div.sphx-glr-download a {
    background-color: #E3F0F6;
    background-image: none;
    color: #11557c;
    border: 0;
}

div.sphx-glr-download a:hover {
    background-color: #BCD4DF;
}

/* Do not fold multiple figures in examples into two column layout. */
img.sphx-glr-multi-img {
    max-width: 100%;
}

table.property-table th,
table.property-table td {
    padding: 4px 10px;
}

/* Fix selection of parameter names; remove when fixed in the theme
 * https://github.com/sphinx-doc/sphinx/pull/9763
 */
.classifier:before {
    display: inline-block;
    margin: 0 0.5em;
}

/* Make the code examples in the API reference index the same height. */
.api-interface-example pre {
    min-height: 6.5rem;
}

/* Make inheritance images have a scroll bar if necessary. */
div.graphviz {
    border: 1px solid lightgrey;
    max-height: 50em;
    overflow: auto;
}
img.graphviz.inheritance {
    max-width: none;
}

/* Make tables in notes horizontally scrollable if too large. */
div.wide-table {
    overflow-x: auto;
}

div.wide-table table th.stub {
    background-color: var(--pst-color-background);
    background-clip: padding-box;
    left: 0;
    position: sticky;
}

.imrot-img {
    display: flex;
    margin: auto;
    max-width:15em;
    align-self: center;
  }

  .imrot-cap {
    text-align: center;
    font-style: italic;
    font-size:  large;
  }


.checklist {
    list-style: none;
    padding: 0;
    margin: 0;
}
.checklist li {
    margin-left: 24px;
    padding-left: 23px;
    margin-right: 6px;
}
.checklist li:before {
    content: "\2610\2001";
    margin-left: -24px;
}
.checklist li p {
    display: inline;
}

/* sdd is a custom class that strips out styling from dropdowns
 * Example usage:
 *
 * .. dropdown::
 *   :class-container: sdd
 *
 */

.sdd.sd-dropdown {
    box-shadow: none!important;
}

.sdd.sd-dropdown.sd-card{
    border-style: solid !important;
    border-color: var(--pst-color-border) !important;
    border-width: thin !important;
    border-radius: .05
}

.sdd.sd-dropdown .sd-card-header{
    --pst-sd-dropdown-color: none;
}

.sdd.sd-dropdown .sd-card-header +.sd-card-body{
    --pst-sd-dropdown-color: none;
}

/* section-toc is a custom class that removes the page title from a toctree listing
 * Example usage:
 *
 * .. rst-class:: section-toc
 * .. toctree::
 *
 */
.section-toc.toctree-wrapper .toctree-l1>a{
    display: none;
}
.section-toc.toctree-wrapper li>ul{
    padding-inline-start:0;
}
